//-----------------------------------------------------------------------------
// MK3_2560_TWI_MASTER
// Version 1.0
// 20.02.2012 / UB
// "www.Mikrocontroller-4U.de"
// mc-4u(@)t-online.de
//-----------------------------------------------------------------------------
// Fuer das myAVR MK3 Board
// mit ATMega-2560 @ 16MHz
// TWI-Master-Funktion
// TWI-Kommunikation zwischen einem Master und bis zu 8 Slaves
//-----------------------------------------------------------------------------
// das File muss im Ordner
// "myAVR/WinAVR/AVR/Include/UB_MK3_2560/"
// liegen und mit #include <UB_MK3_2560/MK3_2560_TWI_MASTER.h>
// eingebunden werden
//-----------------------------------------------------------------------------
// Belegt : TWI-BUS
// Benutzt : MK3_2560_TWI.h = Port-D (Bit0 bis Bit1)
//           MK3_2560_TWI_KOM.h
//-----------------------------------------------------------------------------
// Verbindungen :
// SCL vom Master muss zum SCL-Pin aller Slaves
// SDA vom Master muss zum SDA-Pin aller Slaves
//-----------------------------------------------------------------------------
// Hinweise :
// Das Master-Modul darf auch weitere TWI-Module (z.B. LM75) einbinden
// Es muss fuer SCL und SDA je einen Pull-Up Widerstand geben
//-----------------------------------------------------------------------------


//----------------------------------------------------------------------------
// Includes
//----------------------------------------------------------------------------
#include "MK3_2560_TWI.h" //  bindet die TWI-Funktionen ein
#include "MK3_2560_TWI_KOM.h" 

//-----------------------------------------------------------------------------
// INFOS zur Benutzung :
// 01. Funktion "MK3_INIT_TWI_MASTER" starten
// 02. Funktion "MK3_TWIM_WRITE8_SLAVE" schreibt ein Byte ins RAM von einem Slave
//  -> Rueckgabewert muss >TWI_NO_FAILURE< sein
// 03. Funktion "MK3_TWIM_READ8_SLAVE" liest ein Byte vom RAM eines Slaves
//  -> Rueckgabewert muss >TWI_NO_FAILURE< sein
//  03b. per Funktion "MK3_TWIM_RD8_DATA" das Daten Byte auslesen
// 04. Funktion "MK3_TWIM_WRITE16_SLAVE" schreibt ein INT ins RAM von einem Slave
//  -> Rueckgabewert muss >TWI_NO_FAILURE< sein
// 05. Funktion "MK3_TWIM_READ16_SLAVE" liest ein INT vom RAM eines Slaves
//  -> Rueckgabewert muss >TWI_NO_FAILURE< sein
//  05b. per Funktion "MK3_TWIM_RD16_DATA" die INT Daten auslesen
//-----------------------------------------------------------------------------


//-----------------------------------------------------------------------------
// Enumerationen fuer die Funktioen
//-----------------------------------------------------------------------------
// Rueckgabewerte von "WRITE_SLAVE, "READ_SLAVE"
typedef enum {
  TWI_NO_FAILURE =0,    // Slave hat Kommando akzeptiert und bearbeitet
  TWI_WRONG_DATA,       // ungueltige Daten
  TWI_WRONG_START,      // Start-Kennung falsch
  TWI_NO_ACK            // kein ACK vom Slave empfangen
}TWI_SEND_ID_t;


//-----------------------------------------------------------------------------
// Funktionen die vom MAIN benutzt werden koennen
//-----------------------------------------------------------------------------
void MK3_INIT_TWI_MASTER(void);
TWI_SEND_ID_t MK3_TWIM_WRITE8_SLAVE(TWI_Slave_ADR_t twi_adr, char adr, char wert);
TWI_SEND_ID_t MK3_TWIM_READ8_SLAVE(TWI_Slave_ADR_t twi_adr, char adr);
unsigned char MK3_TWIM_RD8_DATA(void);
TWI_SEND_ID_t MK3_TWIM_WRITE16_SLAVE(TWI_Slave_ADR_t twi_adr, char adr, unsigned int wert);
TWI_SEND_ID_t MK3_TWIM_READ16_SLAVE(TWI_Slave_ADR_t twi_adr, char adr);
unsigned int MK3_TWIM_RD16_DATA(void);
//-----------------------------------------------------------------------------


//-----------------------------------------------------------------------------
// Pausen Einstellungen
// damit keine Uebertragungsfehler durch zu schnelles senden entstehen
// diese Werte muessen eventuell an den langsamsten Slave angepasst werden
//-----------------------------------------------------------------------------

#define TWI_MASTER_PAUSE1  10  // kleine pause nach jedem WRITE
#define TWI_MASTER_PAUSE2  10  // kleine pause nach jedem READ



//-----------------------------------------------------------------------------
// interne Struktur vom Master
//-----------------------------------------------------------------------------
typedef struct {
  unsigned char slave_data_lo; // hier stehen die uebergebene daten vom slave
  unsigned char slave_data_hi; // hier stehen die uebergebene daten vom slave
}TWI_MASTER_t;
TWI_MASTER_t TWI_MASTER;





//----------------------------------------------------------------------
// Initialisiert die CPU als TWI-Master
// WICHTIG : es darf nur einen Master im System geben
//----------------------------------------------------------------------
void MK3_INIT_TWI_MASTER(void) {

  // init vom TWI-BUS
  MK3_INIT_TWI(); 

  // Reset aller Variablen
  TWI_MASTER.slave_data_lo=0;
  TWI_MASTER.slave_data_hi=0;

}


//----------------------------------------------------------------------
// sendet ein byte zum Slave zum speichern im RAM
// -> ruckgabe = ok, bzw. fehlermeldung
//-----------------------------------------------------------------------------
TWI_SEND_ID_t MK3_TWIM_WRITE8_SLAVE(TWI_Slave_ADR_t twi_adr, char adr, char wert) {
  TWI_SEND_ID_t ret_wert=TWI_NO_FAILURE;
  char antwort=0;
  bool ok=false;
  
  // adresse pruefen
  if(adr<MAX_TWI_RAM) {

  }
  else {
    ret_wert=TWI_WRONG_DATA;
    return(ret_wert);    
  }

  ok=MK3_TWI_START();                    // sendet Start-Kennung    
  if(ok==true) {
    // nur wenn Startkennung erfolgreich
    antwort=MK3_TWI_SEND(twi_adr,false); // 1.Byte = Slave-Adresse (ohne ACK)
    if(antwort==TWSR_SLAVE_OK) {
      // Slave hat mit ACK geantwortet
      MK3_TWI_SEND(adr,false);           // 2.Byte = Adresse (ohne ACK)
      MK3_TWI_SEND(wert,false);          // 3.Byte = Wert (ohne ACK)
    }
    else {
      // Slave hat nicht geantwortet 
      ret_wert=TWI_NO_ACK;
    }
    MK3_TWI_STOP();                      // sendet Stop-Kennung
    MK3_TWI_WAIT(TWI_MASTER_PAUSE1);     // kurze Pause
  }
  else {
    // Startkennung falsch
    MK3_TWI_STOP();                      // sendet Stop-Kennung
    ret_wert=TWI_WRONG_START;
  }

  return(ret_wert);
}

//----------------------------------------------------------------------
// liest ein byte vom Slave aus dem RAM
// -> ruckgabe = ok, bzw. fehlermeldung
// -> in slave_data_lo steht bei ok die gelesenen daten vom slave
//----------------------------------------------------------------------
TWI_SEND_ID_t MK3_TWIM_READ8_SLAVE(TWI_Slave_ADR_t twi_adr, char adr) {
  TWI_SEND_ID_t ret_wert=TWI_NO_FAILURE;
  char antwort=0;
  bool ok=false;
  unsigned char wert=0;

  // adresse pruefen
  if(adr<MAX_TWI_RAM) {

  }
  else {
    ret_wert=TWI_WRONG_DATA;
    return(ret_wert);    
  }

  ok=MK3_TWI_START();                    // sendet Start-Kennung    
  if(ok==true) {
    // nur wenn Startkennung erfolgreich
    antwort=MK3_TWI_SEND(twi_adr,false); // 1.Byte = Slave-Adresse (ohne ACK)
    if(antwort==TWSR_SLAVE_OK) {
      // Slave hat mit ACK geantwortet
      MK3_TWI_SEND(adr,false);           // 2.Byte = Adresse (ohne ACK)       
      MK3_TWI_START();                   // sendet Start-Kennung
      MK3_TWI_SEND(twi_adr+1,false);     // 1.Byte = Slave-Adresse+1 (ohne ACK)		
      wert=MK3_TWI_READ(false);          // 2.Byte wird vom Slave gesendet (ohne ACK)	
      // wert speichern
      TWI_MASTER.slave_data_lo=wert;	
    }
    else {
      // Slave hat nicht geantwortet
      ret_wert=TWI_NO_ACK; 
    }
    MK3_TWI_STOP();                      // sendet Stop-Kennung
    MK3_TWI_WAIT(TWI_MASTER_PAUSE2);     // kurze Pause
  }
  else {
    // Startkennung falsch
    MK3_TWI_STOP();                      // sendet Stop-Kennung
    ret_wert=TWI_WRONG_START;
  }

  return(ret_wert);
}


//----------------------------------------------------------------------
// sendet ein Integer zum Slave zum speichern im RAM
// -> ruckgabe = ok, bzw. fehlermeldung
//-----------------------------------------------------------------------------
TWI_SEND_ID_t MK3_TWIM_WRITE16_SLAVE(TWI_Slave_ADR_t twi_adr, char adr, unsigned int wert) {
  TWI_SEND_ID_t ret_wert=TWI_NO_FAILURE;
  char antwort=0;
  bool ok=false;
  unsigned char hi,lo;
  
  // adresse pruefen
  if(adr<MAX_TWI_RAM-1) {

  }
  else {
    ret_wert=TWI_WRONG_DATA;
    return(ret_wert);    
  }

  // Daten aufteilen
  lo=(unsigned char)(wert&0x00FF);    
  hi=(unsigned char)(wert>>8);


  ok=MK3_TWI_START();                    // sendet Start-Kennung    
  if(ok==true) {
    // nur wenn Startkennung erfolgreich
    antwort=MK3_TWI_SEND(twi_adr,false); // 1.Byte = Slave-Adresse (ohne ACK)
    if(antwort==TWSR_SLAVE_OK) {
      // Slave hat mit ACK geantwortet
      MK3_TWI_SEND(adr,false);           // 2.Byte = Adresse (ohne ACK)
      MK3_TWI_SEND(hi,false);            // 3.Byte = Hi-byte (ohne ACK)
      MK3_TWI_SEND(lo,false);            // 4.Byte = Lo-byte (ohne ACK)
    }
    else {
      // Slave hat nicht geantwortet 
      ret_wert=TWI_NO_ACK;
    }
    MK3_TWI_STOP();                      // sendet Stop-Kennung
    MK3_TWI_WAIT(TWI_MASTER_PAUSE1);     // kurze Pause
  }
  else {
    // Startkennung falsch
    MK3_TWI_STOP();                      // sendet Stop-Kennung
    ret_wert=TWI_WRONG_START;
  }

  return(ret_wert);
}


//----------------------------------------------------------------------
// liest ein Integer vom Slave aus dem RAM
// -> ruckgabe = ok, bzw. fehlermeldung
// -> in slave_data_lo und hi steht bei ok die gelesenen daten vom slave
//----------------------------------------------------------------------
TWI_SEND_ID_t MK3_TWIM_READ16_SLAVE(TWI_Slave_ADR_t twi_adr, char adr) {
  TWI_SEND_ID_t ret_wert=TWI_NO_FAILURE;
  char antwort=0;
  bool ok=false;
  unsigned char wert;

  // adresse pruefen
  if(adr<MAX_TWI_RAM-1) {

  }
  else {
    ret_wert=TWI_WRONG_DATA;
    return(ret_wert);    
  }

  ok=MK3_TWI_START();                    // sendet Start-Kennung    
  if(ok==true) {
    // nur wenn Startkennung erfolgreich
    antwort=MK3_TWI_SEND(twi_adr,false); // 1.Byte = Slave-Adresse (ohne ACK)
    if(antwort==TWSR_SLAVE_OK) {
      // Slave hat mit ACK geantwortet
      MK3_TWI_SEND(adr,false);           // 2.Byte = Adresse (ohne ACK)       
      MK3_TWI_START();                   // sendet Start-Kennung
      MK3_TWI_SEND(twi_adr+1,false);     // 1.Byte = Slave-Adresse+1 (ohne ACK)		
      wert=MK3_TWI_READ(true);           // 2.Byte wird vom Slave gesendet (mit ACK)
      TWI_MASTER.slave_data_hi=wert;
      wert=MK3_TWI_READ(false);          // 3.Byte wird vom Slave gesendet (ohne ACK)
      TWI_MASTER.slave_data_lo=wert;		
    }
    else {
      // Slave hat nicht geantwortet
      ret_wert=TWI_NO_ACK; 
    }
    MK3_TWI_STOP();                      // sendet Stop-Kennung
    MK3_TWI_WAIT(TWI_MASTER_PAUSE2);     // kurze Pause
  }
  else {
    // Startkennung falsch
    MK3_TWI_STOP();                      // sendet Stop-Kennung
    ret_wert=TWI_WRONG_START;
  }

  return(ret_wert);
}

//----------------------------------------------------------------------
// uebergibt ein Byte vom zuletzt ausgefuehrten Kommando
//----------------------------------------------------------------------
unsigned char MK3_TWIM_RD8_DATA(void) {
  unsigned char ret_wert=0;
  
  ret_wert=TWI_MASTER.slave_data_lo;
  
  return(ret_wert);
}

//----------------------------------------------------------------------
// uebergibt ein Integer vom zuletzt ausgefuehrten Kommando
//----------------------------------------------------------------------
unsigned int MK3_TWIM_RD16_DATA(void) {
  unsigned int ret_wert=0;
  unsigned char lo,hi;
  
  lo=TWI_MASTER.slave_data_lo;
  hi=TWI_MASTER.slave_data_hi;
  
  ret_wert=(hi<<8);
  ret_wert=(ret_wert&0xFF00);
  ret_wert+=lo;   
  
  return(ret_wert);
}



