//-----------------------------------------------------------------------------
// MK3_2560_TWI_SLAVE
// Version 1.0
// 20.02.2012 / UB
// "www.Mikrocontroller-4U.de"
// mc-4u(@)t-online.de
//-----------------------------------------------------------------------------
// Fuer das myAVR MK3 Board
// mit ATMega-2560 @ 16MHz
// TWI-Slave-Funktion
// TWI-Kommunikation zwischen einem Master und bis zu 8 Slaves
//-----------------------------------------------------------------------------
// das File muss im Ordner
// "myAVR/WinAVR/AVR/Include/UB_MK3_2560/"
// liegen und mit #include <UB_MK3_2560/MK3_2560_TWI_SLAVE.h>
// eingebunden werden
//-----------------------------------------------------------------------------
// Belegt : TWI-BUS Port-D (Bit0 bis Bit1)
// Benutzt : MK3_2560_TWI_KOM.h 
//-----------------------------------------------------------------------------
// Verbindungen :
// SCL vom Slave muss zum SCL-Pin vom Master
// SDA vom Slave muss zum SDA-Pin vom Master
//-----------------------------------------------------------------------------
// Hinweise :
// Das Slave-Modul darf keine weiteren TWI-Module (z.B. LM75) einbinden
// Es muss fuer SCL und SDA je einen Pull-Up Widerstand geben
//-----------------------------------------------------------------------------
// TWI-Funktionen :
// - alle TWI-Funktionen gehen vom Master aus
// - der TWI-CPU-Slave verhaellt sich wie ein TWI-RAM mit 256 Bytes speicher
// - der Adress-Pointer wird automatisch incrementiert
//   d.H. es reicht beim lesen und schreiben die Startadresse zu senden
//   und dann beliebig viele Daten hintereinander zu schreiben oder zu lesen
//   
//   WRITE [1 Byte]  : START|SLAVE_ADR|RAM_ADR|D0|STOP
//   WRITE [4 Bytes] : START|SLAVE_ADR|RAM_ADR|D0|D1|D2|D3|STOP
//     -> beim WRITE muessen alle Bytes ohne ACK gesendet werden 
//
//   READ [1 Byte]   : START|SLAVE_ADR|RAM_ADR|START|SLAVE_ADR+1|D0|STOP
//   READ [4 Bytes]  : START|SLAVE_ADR|RAM_ADR|START|SLAVE_ADR+1|D0|D1|D2|D3|STOP
//     -> beim READ muss das letzte Byte ohne ACK,
//        alle anderen mit ACK gelesen werden
//-----------------------------------------------------------------------------


#ifndef MK3_2560_TWI    // Wichtig !!
#define MK3_2560_TWI 1  // damit werden die TWI-Master Funktionen unterbunden



//----------------------------------------------------------------------------
// Includes
//----------------------------------------------------------------------------
#include "MK3_2560_TWI_KOM.h" 
#include <avr/interrupt.h>




//-----------------------------------------------------------------------------
// INFOS zur Benutzung :
// 01. Funktion "MK3_INIT_TWI_SLAVE" starten
//   -> die komplette TWI-Kommunikation laeuft im Hintergrund per Interrupt ab
// 02. Funktion "MK3_TWIS_AKTIVITY" prueft ob es TWI aktivitaet gab
//   02a : falls Ja -> ist der Reckgabewert != KOM_TWI_NONE
// 03. Funktion "MK3_TWIS_WR8_RAM" schreibt ein Byte ins eigene RAM
// 04. Funktion "MK3_TWIS_RD8_RAM" liest ein Byte aus eigenem RAM
// 05. Funktion "MK3_TWIS_WR16_RAM" schreibt ein INT ins eigene RAM
// 06. Funktion "MK3_TWIS_RD16_RAM" liest ein INT aus eigenem RAM
//-----------------------------------------------------------------------------



//-----------------------------------------------------------------------------
// Enumerationen fuer die Funktioen
//-----------------------------------------------------------------------------
// Rueckgabewerte von "MK3_TWIS_AKTIVITY"
typedef enum {
  KOM_TWI_NONE =0,  // es hat keine neue Kommunikation stattgefunden
  KOM_TWI_RD,       // es wurde ein RAM-Read Kommando ausgefuehrt
  KOM_TWI_WR,       // es wurde ein RAM-Write Kommando ausgefuehrt
  KOM_TWI_ERR       // es gab einen Kommando fehler
}TWI_KOM_STATUS_t;


//-----------------------------------------------------------------------------
// Funktionen die vom MAIN benutzt werden koennen
//-----------------------------------------------------------------------------
void MK3_INIT_TWI_SLAVE(TWI_Slave_ADR_t twi_adr);     // zum initialis. vom TWI als Slave
TWI_KOM_STATUS_t MK3_TWIS_AKTIVITY(void);             // erkennt TWI aktivitaeten
void MK3_TWIS_WR8_RAM(char adr, char daten);          // schreibt ein byte ins RAM
unsigned char MK3_TWIS_RD8_RAM(char adr);             // liest ein byte aus dem RAM
void MK3_TWIS_WR16_RAM(char adr, unsigned int daten); // schreibt ein INT ins RAM
unsigned int MK3_TWIS_RD16_RAM(char adr);             // liest ein INT aus dem RAM
//-----------------------------------------------------------------------------


//-----------------------------------------------------------------------------
// Globale Variabeln
//-----------------------------------------------------------------------------
// keine



//-----------------------------------------------------------------------------
// Port-Pin definition
//-----------------------------------------------------------------------------
#define MK3_DDR_TWI DDRD	// DDR-Register des TWI = PORT-D

#define MK3_TWI_SCL 0x01    	// CLOCK von TWI = Bit0
#define MK3_TWI_SDA 0x02        // DATA von TWI  = Bit1


//-----------------------------------------------------------------------------
// Register definition TWCR
//-----------------------------------------------------------------------------
#define BIT_TWIE  0x01          // TWIE  = BIT-0
#define BIT_TWEN  0x04          // TWEN  = BIT-2
#define BIT_TWSTO 0x10          // TESTO = BIT-4
#define BIT_TWEA  0x40          // TWEA  = BIT-6
#define BIT_TWINT 0x80          // TWINT = BIT-7

//-----------------------------------------------------------------------------
// Register definition TWSR
//-----------------------------------------------------------------------------
#define MASKE_TWSR     0xF8     // Bit7 bis Bit3
#define TWSR_ERROR     0x00     // TWI-Error
#define TWSR_OWN_ADR_W 0x60     // Eigene TWI-Adresse empfangen (Write)
#define TWSR_DATA      0x80     // Daten wurden empfangen
#define TWSR_STOP      0xA0     // Stop wurde empfangen
#define TWSR_OWN_ADR_R 0xA8     // Eigene TWI-Adresse empfangen (Read)
#define TWSR_READ_ACK  0xB8     // Read Befehl mit ACK
#define TWSR_READ_NACK 0xC0     // Read Befehl mit NACK (letztes Byte)


//-----------------------------------------------------------------------------
// interne Struktur vom Slave
//-----------------------------------------------------------------------------
typedef enum {
  TWI_WAIT   =0,   // Warte auf TWI-Kommando
  TWI_START,       // Start empfangen
  TWI_ADR,         // Adresse empfangen
  TWI_WRITE,       // Daten ins RAM schreiben
  TWI_READ         // Daten vom RAM lesen
}TWI_State_t;

typedef struct {
  TWI_State_t twi_state;           // State-Machine  
  TWI_KOM_STATUS_t last_status;    // letzter status
  unsigned char akt_ram_ptr;       // pointer auf das RAM
  unsigned char ram[MAX_TWI_RAM];  // RAM fuer Datenaustausch mit Master
}TWI_SLAVE_t;
TWI_SLAVE_t TWI_SLAVE;



//-----------------------------------------------------------------------------
// Globale interne Variablen
//-----------------------------------------------------------------------------
bool twi_init=false;

//----------------------------------------------------------------------
// Initialisiert die CPU als TWI-Slave
// mit einer bestimmten Slave-Adresse
// WICHTIG : die Adresse darf nur einmal im System benutzt werden
//----------------------------------------------------------------------
void MK3_INIT_TWI_SLAVE(TWI_Slave_ADR_t twi_adr) {
  unsigned int n;

  if(twi_init==false) {
    // die initialisierung wird nur einmal gemacht
    twi_init=true;

    // init der Port-Pins eigentlich nicht notwendig  
    MK3_DDR_TWI&=~MK3_TWI_SCL;	// SCL als Eingang
    MK3_DDR_TWI&=~MK3_TWI_SDA;	// SDA als Eingang

    // Einstellungen fuer Slave
    TWBR=0x00;                  // TWI-Clock wird vom Master generiert
    TWCR=0x00;
    TWSR=0x00;
    TWDR=0xFF;                  // dummy Inhalt

    // TWI-Slave Adresse setzen
    TWAR=(twi_adr & 0xFE);      // nur 7 bit setzen

    // TWI einschalten (als Slave-Receiver)
    TWCR|=BIT_TWEN;		// TWI enable 
    TWCR|=BIT_TWEA;		// Acknoledge setzen
    TWCR|=BIT_TWIE;             // TWI-Interrupt aktivieren
    TWCR|=BIT_TWINT;		// Interrupt Flag loeschen

    // Reset der Variablen------------------------------
    TWI_SLAVE.twi_state=TWI_WAIT;    
    TWI_SLAVE.last_status=KOM_TWI_NONE;
    TWI_SLAVE.akt_ram_ptr=0;

    // init vom RAM-------------------------------------
    for(n=0;n<MAX_TWI_RAM;n++) {
      TWI_SLAVE.ram[n]=0x00;
    }

    SREG|=0x80; // Alle Interrupts einschalten
  }  
}

//----------------------------------------------------------------------
// rueckgabe welches TWI Kommando bearbeitet wurde
//----------------------------------------------------------------------
TWI_KOM_STATUS_t MK3_TWIS_AKTIVITY(void) {
  TWI_KOM_STATUS_t ret_wert=KOM_TWI_NONE;

  if(TWI_SLAVE.last_status!=KOM_TWI_NONE) {
    // falls eine aktivitaet aufgetreten ist
    ret_wert=TWI_SLAVE.last_status;
    TWI_SLAVE.last_status=KOM_TWI_NONE;
  }

  return(ret_wert);
}

//----------------------------------------------------------------------
// schreibt ein Byte ins RAM
//----------------------------------------------------------------------
void MK3_TWIS_WR8_RAM(char adr, char daten) {
  unsigned int n=0;

  n=adr;
  
  if(n<MAX_TWI_RAM) {
    TWI_SLAVE.ram[n]=daten;
  }
}

//----------------------------------------------------------------------
// liest ein Byte aus dem RAM
//----------------------------------------------------------------------
unsigned char MK3_TWIS_RD8_RAM(char adr) {
  unsigned char ret_wert=0;
  unsigned int n=0;

  n=adr;

  if(n<MAX_TWI_RAM) {
    ret_wert=TWI_SLAVE.ram[n];
  }
  
  return(ret_wert);
}

//----------------------------------------------------------------------
// schreibt ein INT ins RAM
//----------------------------------------------------------------------
void MK3_TWIS_WR16_RAM(char adr, unsigned int daten) {
  unsigned char lo,hi; 
  unsigned int n=0;
  
  n=adr; 
  
  if(n<MAX_TWI_RAM-1) {
    lo=(unsigned char)(daten&0x00FF);    
    hi=(unsigned char)(daten>>8);
    
    TWI_SLAVE.ram[n+0]=hi;
    TWI_SLAVE.ram[n+1]=lo;
  }
}

//----------------------------------------------------------------------
// liest ein INT aus dem RAM
//----------------------------------------------------------------------
unsigned int MK3_TWIS_RD16_RAM(char adr) {
  unsigned int ret_wert=0;
  unsigned char lo,hi;
  unsigned int n=0;

  n=adr;

  if(n<MAX_TWI_RAM-1) {
    hi=TWI_SLAVE.ram[n+0];
    lo=TWI_SLAVE.ram[n+1];
     
    ret_wert=(hi<<8);
    ret_wert=(ret_wert&0xFF00);
    ret_wert+=lo;
  }
  
  return(ret_wert);
}



//-----------------------------------------------------------------------------
ISR (TWI_vect) {                          // ISR : TWI
  unsigned char status,wert;

  status=(TWSR&MASKE_TWSR);   // TWSR-Register auslesen
  wert=TWDR; // TWDR-Register auslesen  

  // Status-Byte auswerten
  switch(status) {
    case TWSR_OWN_ADR_W : // eigene TWI-Adresse wurde empfangen (Write) (0x60)
      if(TWI_SLAVE.twi_state==TWI_WAIT) {	    
        TWI_SLAVE.twi_state=TWI_START; // auf START weiterschalten				        
      }	  	    
    ;break;
    case TWSR_DATA : // Daten wurden empfangen (0x80)
      if(TWI_SLAVE.twi_state==TWI_START) {
        // als erstes den RAM Pointer setzen		        
        // RAM Grenze pruefen
        if(wert<MAX_TWI_RAM) {
          // OK
          TWI_SLAVE.akt_ram_ptr=wert;
          TWI_SLAVE.twi_state=TWI_ADR; // auf ADR weiterschalten
        }
        else {
          // Fehler
          TWI_SLAVE.akt_ram_ptr=0;
          TWI_SLAVE.twi_state=TWI_WAIT; // auf WAIT zurueckschalten
        }
      }
      else if(TWI_SLAVE.twi_state==TWI_ADR) {
        // alle anderen daten im RAM speichern
        TWI_SLAVE.twi_state=TWI_WRITE; // auf WRITE weiterschalten
        TWI_SLAVE.ram[TWI_SLAVE.akt_ram_ptr]=wert;
        // RAM Grenze pruefen
        if(TWI_SLAVE.akt_ram_ptr<(MAX_TWI_RAM-1)) {
          // naechste RAM Adresse
          TWI_SLAVE.akt_ram_ptr++;
        }
      }
      else if(TWI_SLAVE.twi_state==TWI_WRITE) {
        // alle anderen daten im RAM speichern
        TWI_SLAVE.ram[TWI_SLAVE.akt_ram_ptr]=wert;
        // RAM Grenze pruefen
        if(TWI_SLAVE.akt_ram_ptr<(MAX_TWI_RAM-1)) {
          // naechste RAM Adresse
          TWI_SLAVE.akt_ram_ptr++;
        }        
      }   	  
    ;break;	
    case TWSR_OWN_ADR_R : // eigene TWI-Adresse wurde empfangen (Read) (0xA8)
      if(TWI_SLAVE.twi_state==TWI_WAIT) {
        TWI_SLAVE.twi_state=TWI_READ; // auf READ weiterschalten
        // RAM inhalt beim naechsten READ senden
        TWDR=TWI_SLAVE.ram[TWI_SLAVE.akt_ram_ptr];
        // RAM Grenze pruefen
        if(TWI_SLAVE.akt_ram_ptr<(MAX_TWI_RAM-1)) {
          // naechste RAM Adresse
          TWI_SLAVE.akt_ram_ptr++;
        }
      }
    ;break;
    case TWSR_READ_ACK : // Read Befehl mit ACK (0xB8)
      if(TWI_SLAVE.twi_state==TWI_READ) {
        // RAM inhalt beim naechsten READ senden
        TWDR=TWI_SLAVE.ram[TWI_SLAVE.akt_ram_ptr];
        // RAM Grenze pruefen
        if(TWI_SLAVE.akt_ram_ptr<(MAX_TWI_RAM-1)) {
          // naechste RAM Adresse
          TWI_SLAVE.akt_ram_ptr++;
        }	  
      }      
    ;break;
    case TWSR_READ_NACK : // Read Befehl mit NACK (letztes Byte) (0xC0)
      // letztes Byte fuer READ gesendet
      if(TWI_SLAVE.twi_state==TWI_READ) {
        // aktuellen Status speichern
        TWI_SLAVE.last_status=KOM_TWI_RD;
      } 
      TWI_SLAVE.twi_state=TWI_WAIT; // auf WAIT zurueckschalten
    ;break;
    case TWSR_STOP : // STOP wurde empfangen (0xA0)
      if(TWI_SLAVE.twi_state==TWI_WRITE) {
        // aktuellen Status speichern
        TWI_SLAVE.last_status=KOM_TWI_WR;
      }       
      TWI_SLAVE.twi_state=TWI_WAIT; // auf WAIT zurueckschalten    
    ;break;
    case TWSR_ERROR : // TWI-Error (0x00)
      TWCR|=BIT_TWSTO;  // um den Fehler aufzuloesen
      TWI_SLAVE.twi_state=TWI_WAIT; // auf WAIT zurueckschalten
      TWI_SLAVE.last_status=KOM_TWI_ERR;
    ;break;
    default : // bei allen anderen Werten von TWSR
      TWI_SLAVE.twi_state=TWI_WAIT; // auf WAIT zurueckschalten
      TWI_SLAVE.last_status=KOM_TWI_ERR;       
  }

  TWCR|=BIT_TWINT; // Interrupt Flag loeschen
}



#else
  // falls der Compiler hier landet, wurde ein anderes
  // TWI-Modul included...das ist bei einem TWI-Slave nicht erlaubt
  # warning "EIN TWI-Slave darf kein TWI-Modul benutzen"  
#endif  // Ende von #define MK3_2560_TWI



